# std
import os
import subprocess
import shutil
import mimetypes
# pip packages
from pymediainfo import MediaInfo


def anti_file_prefix(path: str) -> str:
    """Remove prefixed 'file://' from path"""
    if path.lower().startswith("file://"):
        path = path[7:]
        if os.name == "nt":
            # windows sometimes adds an extra leading /
            path = path.lstrip("/")
    return path


def get_mime_type(file_path: str) -> str:
    """Get file mime-type based on file contents or extension"""
    # initialise mime-types, let it load all mimes
    mimetypes.init()
    # get the file extension
    file_ext = os.path.splitext(file_path)[-1]
    # check for special file types if theres no mime type
    if file_ext not in mimetypes.types_map:
        # check if the file is a D2V/DGIndexProjectFile
        with open(file_path, mode="rb") as f:
            if f.read(18) == "DGIndexProjectFile".encode("utf-8"):
                if f.read(2) != bytes([0x31, 0x36]):
                    raise ValueError(
                        "pvsfunc.get_file_type: D2V was created with an unsupported indexer, please use DGIndex v1.5.8." +
                        (" It works perfectly fine under Wine." if os.name != "nt" else "")
                    )
                return "video/d2v"
        if file_ext.lower() == ".vob":
            return "video/vob"
        raise ValueError(f"pvsfunc.get_file_type: Unrecognised file extension ({file_ext})")
    mime_type = mimetypes.types_map[file_ext] if file_ext in mimetypes.types_map else None
    # ensure that the mime is a video or image file
    if not mime_type.startswith("video/") and not mime_type.startswith("image/"):
        raise ValueError(f"pvsfunc.get_file_type: Only Video or Image files are supported. ({mime_type})")
    # return the mime
    return mime_type


def get_video_codec(file_path: str) -> str:
    """Get video codec using MediaInfo"""
    if not os.path.exists(file_path) or not os.path.isfile(file_path):
        return "?"
    track = [t for t in MediaInfo.parse(
        filename=file_path
    ).tracks if t.track_type in ["Video", "Image"]]
    if not track:
        raise ValueError("No video/image track in file...")
    track = track[0]
    # we try both as in some cases codec_id isn't set
    codec = track.codec_id or track.commercial_name
    # do some squashing to reduce amount of code repetition
    if codec == "MPEG-1 Video":
        return "V_MPEG1"
    if codec == "MPEG-2 Video":
        return "V_MPEG2"
    return codec


def get_d2v(file_path: str) -> str:
    """Demux video track and generate a D2V file for it if needed"""
    IS_VOB = os.path.splitext(file_path)[-1].lower() == ".vob"
    # create file_path location of the d2v path
    d2v_path = f"{os.path.splitext(file_path)[0]}.d2v"
    if os.path.exists(d2v_path):
        print("Skipping generation as a D2V file already exists")
        return d2v_path
    # demux the mpeg stream if needed
    vid_path = file_path
    if not IS_VOB:
        if os.path.splitext(file_path)[-1].lower() != ".mpeg":
            vid_path = f"{os.path.splitext(file_path)[0]}.mpg"
        if os.path.exists(vid_path):
            print("Skipping demuxing of raw mpeg stream as it already exists")
        else:
            mkvextract_path = shutil.which("mkvextract")
            if not mkvextract_path:
                raise RuntimeError(
                    "pvsfunc.PSourcer: Required binary 'mkvextract' not found. "
                    "Install MKVToolNix and make sure it's binaries are in the environment path."
                )
            subprocess.run([
                mkvextract_path, os.path.basename(file_path),
                # todo ; this assumes the track with track-id of 0 is the video, not ideal
                "tracks", f"0:{os.path.basename(vid_path)}"
            ], cwd=os.path.dirname(file_path))
    # use dgindex to create a d2v file for the demuxed track
    dgindex_path = shutil.which("DGIndex.exe") or shutil.which("dgindex.exe")
    if not dgindex_path:
        raise RuntimeError(
            "pvsfunc.PSourcer: This video file will need a required binary 'DGIndex.exe' which isn't found.\n"
            "tl-dr; add DGIndex.exe to your system path. Ensure the executable is named exactly `DGIndex.exe`.\n"
            "Windows: Start Menu -> Environment Variables, Add DGIndex's folder to `PATH` variable.\n"
            "Linux: append to $PATH in /etc/profile, I recommend using `nano /etc/profile.d/env.sh`. Must reboot."
        )
    subprocess.run([
        dgindex_path,
        "-ai" if IS_VOB else "-i", os.path.basename(vid_path),
        "-ia", "5",  # iDCT Algorithm, 5=IEEE-1180 Reference
        "-fo", "2",  # Field Operation, 2=Ignore Pulldown Flags
        "-yr", "1",  # YUV->RGB, 1=PC Scale
        "-om", "0",  # Output Method, 0=None (just d2v)
        "-hide", "-exit",  # start hidden and exit when saved
        "-o", os.path.splitext(os.path.basename(file_path))[0]
    ], cwd=os.path.dirname(file_path))
    # return file path of the new d2v file
    return d2v_path


def calculate_aspect_ratio(width: int, height: int) -> str:
    """Calculate the aspect-ratio gcd string from resolution"""
    def gcd(a, b):
        """The GCD (greatest common divisor) is the highest number that evenly divides both width and height."""
        return a if b == 0 else gcd(b, a % b)
    r = gcd(width, height)
    return f"{int(width / r)}:{int(height / r)}"



# std vs
from vapoursynth import core
import vapoursynth as vs
import os
import functools
# vs func repos
try:
    import havsfunc
except ImportError:
    raise RuntimeError(
        "pvsfunc.PDeinterlacer: Required script havsfunc not found. "
        "https://github.com/HomeOfVapourSynthEvolution/havsfunc"
    )

# pip packages
from pyd2v import D2V


class PDeinterlacer:
    """
    PDeinterlacer (PHOENiX Deinterlacer)
    Deinterlaces a clip with the most optimal wrapping based on the sourcer.
    The clip will need to be loaded from PSourcer to work as it needs it's Props.
    """

    def __init__(self, clip, tff=True, kernel=None, kernel_args=None, debug=False):
        self.clip = clip
        self.tff = tff
        self.kernel = kernel
        self.kernel_args = kernel_args
        self.debug = debug
        # validate arguments
        if not isinstance(self.clip, vs.VideoNode):
            raise TypeError("pvsfunc.PDeinterlacer: This is not a clip")
        # set default kernel to QTGMC
        if not self.kernel:
            self.kernel = havsfunc.QTGMC
        # if kernel is QTGMC, set it's defaults
        if self.kernel == havsfunc.QTGMC:
            self.kernel_args = {
                # defaults
                **{
                    "FPSDivisor": 2,
                    "Preset": "Placebo",
                    "MatchPreset": "Placebo",
                    "MatchPreset2": "Placebo",
                    "TFF": self.tff,
                    "InputType": 0,
                    "SourceMatch": 3,
                    "Lossless": 2,
                    "Sharpness": 0.2,
                    "ShutterBlur": 0,
                    "ShutterAngleSrc": 0,
                    "ShutterAngleOut": 0,
                    "SBlurLimit": 0
                },
                # user configuration
                **dict(self.kernel_args or {})
            }
        self.props = self.clip.get_frame(0).props
        self.props = {k: v.decode("utf-8") if type(v) == bytes else v for k, v in self.props.items()}
        if self.props["PVSSourcer"] == "core.d2v.Source":
            self._d2v()
        elif self.props["PVSSourcer"] == "core.ffms2.Source":
            if kernel == havsfunc.QTGMC:
                kernel_args["FPSDivisor"] = 2  # only supporting same-rate fps atm
            self._ffms2()
        elif self.props["PVSSourcer"] == "core.imwri.Read":
            print("pvsfunc.PDeinterlacer: Warning: This source is a clip of images and cannot be deinterlaced.")
        else:
            raise ValueError(f"Unimplemented deinterlacer for Sourcer {self.props['PVSSourcer']}")
    
    def _d2v(self):
        """
        Very accurate deinterlacing using raw frame metadata to know what to
        deinterlace when necessary. It even fixes the frame rates of progressive
        streams and converts VFR to CFR when necessary.

        For MPEG2, this is as good as it gets in terms of using a deinterlacer.
        """
        # Get D2V object
        self.d2v = D2V(self.props["PVSFilePath"])
        # Get every frames' flag data, this contains information on displaying frames
        flags = [[dict(**y, vob=x["vob"], cell=x["cell"]) for y in x["flags"]] for x in self.d2v.data]
        flags = [f for l in flags for f in l]
        # Get percentage of progressive frames
        progressive_percent = (sum(1 for x in flags if x["progressive_frame"]) / len(flags))*100
        # Get pulldown information
        pulldown_frames = [n for n,f in enumerate(flags) if f["progressive_frame"] and f["rff"] and f["tff"]]
        # todo ; get an mpeg2 that uses Pulldown metadata (rff flags) that ISN'T Pulldown 2:3 to test math
        #        this math seems pretty far fetched, if we can somehow obtain the Pulldown x:x:...
        #        string that mediainfo can get, then calculating it can be much easier and more efficient.
        pulldown_cycle = [n for n,f in enumerate(flags) if f["tff"] and f["rff"]]
        if pulldown_cycle:
            pulldown_cycle = list(zip(pulldown_cycle[::2], pulldown_cycle[1::2]))
            pulldown_cycle = [r - l for l,r in pulldown_cycle]
            pulldown_cycle = max(set(pulldown_cycle), key=pulldown_cycle.count) + 1  # most common entry + 1
        else:
            pulldown_cycle = None

        if progressive_percent != 100.0:
            # video is not all progressive content, meaning it is either:
            # - entirely interlaced
            # - mix of progressive and interlaced sections
            # 1. fix the frame rate of the progressive sections by applying it's pulldown (without interlacing)
            if pulldown_frames:
                self.clip = core.std.DuplicateFrames(clip=self.clip, frames=pulldown_frames)
            # 2. also fix the frame rate of the flag list to match the fixed clip
            pulldown_flags = []
            for flag in flags:
                pulldown_flags.append(flag)
                if flag["progressive_frame"] and flag["rff"] and flag["tff"]:
                    pulldown_flags.append(
                        dict(**{**flag, **{"progressive_frame": True, "rff": False, "tff": False}})
                    )
            # 3. create a clip from the output of the kernel deinterlacer
            deinterlaced_clip = self.kernel(self.clip, **self.kernel_args)
            fps_factor = (deinterlaced_clip.fps.numerator / deinterlaced_clip.fps.denominator)
            fps_factor = fps_factor / (self.clip.fps.numerator / self.clip.fps.denominator)
            if fps_factor != 1.0 and fps_factor != 2.0:
                raise ValueError(
                    f"pvsfunc.PDeinterlacer: The deinterlacer kernel returned an unsupported frame-rate ({deinterlaced_clip.fps}). "
                    "Only single-rate and double-rate is supported with PDeinterlacer at the moment."
                )
            # 4. create a format clip, used for metadata of final clip
            format_clip = core.std.BlankClip(
                clip=deinterlaced_clip,
                length=len(pulldown_flags) * fps_factor
            )
            # 5. deinterlace whats interlaced
            def _d(n, f, c, d, fl, ff):
                flag = fl[int(n / ff)]
                if flag["progressive_frame"]:
                    # progressive frame, we don't need to do any deinterlacing to this frame
                    # though we may need to duplicate it if double-rate fps output
                    rc = core.std.Interleave([c] * ff) if ff > 1 else c
                    return core.text.Text(
                        rc,
                        f"\n\n\n\n\n\n VOB: {flag['vob']}:{flag['cell']} - Frame #{n} - Untouched ",
                        alignment=7
                    ) if self.debug else rc
                # interlaced frame, we need to use `d` (deinterlaced) frame.
                return core.text.Text(
                    d,
                    f"\n\n\n\n\n\n VOB: {flag['vob']}:{flag['cell']} - Frame #{n} - Deinterlaced! ",
                    alignment=7
                ) if self.debug else d
            self.clip = core.std.FrameEval(
                format_clip,
                functools.partial(
                    _d,
                    c=self.clip,
                    d=deinterlaced_clip,
                    fl=pulldown_flags,
                    ff=fps_factor
                ),
                prop_src=self.clip
            )
        else:
            # video is entirely progressive without a hint of interlacing in sight
            # however, it needs it's FPS to be fixed. rff=False with core.d2v.Source
            # resulted in it returning with the FPS set to 30000/1001, let's revert that
            # back to whatever it should be based on its pulldown cycle
            if pulldown_cycle:
                self.clip = core.std.AssumeFPS(
                    self.clip,
                    fpsnum=self.clip.fps.numerator - (self.clip.fps.numerator / pulldown_cycle),
                    fpsden=self.clip.fps.denominator
                )
            if self.debug:
                self.clip = core.std.FrameEval(
                    self.clip,
                    functools.partial(
                        lambda n, f, c, fl: core.text.Text(
                            c,
                            f"\n\n\n\n\n\n VOB: {fl[n]['vob']}:{fl[n]['cell']} - Frame #{n} - Untouched ",
                            alignment=7
                        ),
                        c=self.clip,
                        fl=flags
                    ),
                    prop_src=self.clip
                )
        
        if self.debug:
            fps = self.clip.fps
            if self.clip.fps.numerator == 25:
                fps = "PAL"
            elif self.clip.fps.numerator == 30000:
                fps = "NTSC"
            elif self.clip.fps.numerator == 24:
                fps = "FILM"
            self.clip = core.text.Text(
                self.clip,
                " " + (" \n ".join([
                    f"{os.path.basename(self.props['PVSFilePath'])}",
                    f"{fps}, Loaded with {self.props['PVSSourcer']}",
                    f"- {len(flags):,} coded pictures, which {progressive_percent:.2f}% of are Progressive",
                    f"- {len(pulldown_frames):,} frames are asking for pulldown{f' which occurs every {pulldown_cycle} frames' if pulldown_cycle else ''}",
                    f"- {len(flags) + len(pulldown_frames):,} total frames after pulldown flags are honored"
                ])) + " ",
                alignment=7
            )
    
    def _ffms2(self):
        """
        Deinterlace using ffms2 (ffmpeg) using a basic FieldBased!=0 => QTGMC method
        """
        self.clip = core.std.FrameEval(
            self.clip,
            functools.partial(
                lambda n, f, c, d: (
                    core.text.Text(c, "Untouched Frame (_FieldBased=0)", alignment=1) if self.debug else c
                ) if f.props["_FieldBased"] == 0 else (
                    core.text.Text(d, "Deinterlaced Frame (via QTGMC)", alignment=1) if self.debug else d
                ),
                c=self.clip,
                d=self.kernel(self.clip, **self.kernel_args)
            ),
            prop_src=self.clip
        )



# std vs
from vapoursynth import core
import vapoursynth as vs
# std py
import functools


def decimate(clip, mode=0, cycle=5, offsets=[0, 1, 3, 4], debug=False):
    """
    IVTC (Inverse-telecine) the clip using decimation (frame deletion).
    This would commonly be used to revert the telecine process of FILM
    to NTSC but can be used for other rate changes.

    :param clip: VapourSynth Clip (VideoNode) to decimate
    :param mode: 0=core.std.SelectEvery (recommended), 1=core.vivtc.VDecimate (be warned; its inaccurate!)
    :param cycle: Defines the amount of frames to calculate offsets on at a time.
    :param offsets: Mode 0's offsets zero-indexed list. This indicates which frames to KEEP from the cycle.
    :param debug: Skip decimation and print debugging information. Useful to check if the frames that the cycle
    and offset settings you have provided are correct and actually decimate the right frames.
    """
    if mode < 0 or mode > 1:
        raise ValueError(f"pvsfunc.decimate: Incorrect mode ({mode}), it must be an int value between 0-1")
    if mode == 0:
        if debug:
            return core.std.FrameEval(
                clip,
                functools.partial(
                    lambda n, f, c: core.text.Text(
                        c,
                        f" mode={mode} cycle={cycle} offsets={offsets} \n"
                        f"decimated_frame={(n % cycle) not in offsets}\n",
                        alignment=1
                    ),
                    c=clip
                ),
                prop_src=clip
            )
        return core.std.SelectEvery(clip, cycle=cycle, offsets=offsets)
    if mode == 1:
        if debug:
            return core.std.FrameEval(
                clip,
                functools.partial(
                    lambda n, f, c: core.text.Text(
                        c,
                        f" mode={mode} cycle={cycle} \n"
                        "Important: Please consider another mode. More information: git.io/avoid-tdecimate.\n"
                        f"decimated_frame={f.props['VDecimateDrop'] == 1}\n",
                        alignment=1
                    ),
                    c=clip
                ),
                prop_src=core.vivtc.VDecimate(clip, cycle=cycle, dryrun=True)
            )
        return core.vivtc.VDecimate(clip, cycle=cycle)


def debox(clip, aspect_ratio, mode=0, offset=0):
    """
    Remove Pillarboxing, Letterboxing or Windowboxing from the video
    by calculating a crop area based on `aspect_ratio` calculated against
    clip width and height. If it's Windowboxing, use this function twice,
    first for Pillarboxing, then for Letterboxing.

    :param clip: VapourSynth Clip (VideoNode) to debox
    :param aspect_ratio: The Aspect Ratio you wish to crop to, e.g. `"4:3"` to crop to 4:3
    :param mode: The Direction you wish to crop. `0`=Pillarboxing, `1`=Letterboxing.
    :param offset: If the content isn't exactly in the center of the frame
    """
    aspect_ratio = [int(n) for n in aspect_ratio.split(":")]
    aspect_ratio = aspect_ratio[0] / aspect_ratio[1]
    area = (clip.width - (clip.height * aspect_ratio)) / 2
    return core.std.CropRel(
        clip,
        left=area + offset if mode == 0 else 0,
        top=0 if mode == 0 else area + offset,
        right=area - offset if mode == 0 else 0,
        bottom=0 if mode == 0 else area - offset
    )



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="pvsfunc",
    version="3.2.1",
    author="PHOENiX",
    author_email="pragma.exe@gmail.com",
    description="PHOENiX's compilation of VapourSynth Script's and Functions",
    license="GPLv3",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/rlaPHOENiX/pvsfunc",
    packages=find_packages(),
    install_requires=[
        "vapoursynth",
        "pymediainfo",
        "pyd2v>=1.0.3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# std vs
from vapoursynth import core
import vapoursynth as vs
# pvsfunc
from pvsfunc.helpers import anti_file_prefix, get_mime_type, get_video_codec, get_d2v


CODEC_SOURCER_MAP = {
    "IMAGE": "core.imwri.Read",
    "V_MPEG1": "core.d2v.Source",
    "V_MPEG2": "core.d2v.Source",
    # codecs not listed here will default to `core.ffms2.Source`
}

SOURCER_ARGS_MAP = {
    "core.imwri.Read": {
        # disable the alpha channel as it often causes an incompatibility
        # with some functions due to the extra channel
        "alpha": False
    },
    "core.d2v.Source": {
        # ignore rff (pulldown flags) as we do not want it to interlace
        # the pulldown frames and end up giving us more work...
        "rff": False
    },
    "core.ffms2.Source": {
        # disable the alpha channel as it often causes an incompatibility
        # with some functions due to the extra channel
        "alpha": False
    }
}


class PSourcer:
    """
    PSourcer (PHOENiX Sourcer)
    Loads an input file path with the most optimal clip source based on the file.
    This is mainly a wrapper function for my other classes and functions of pvsfunc.
    The sourcer (and it's arguments) are based on my own personal opinions.
    -
    core.d2v.Source is used on videos that are supported by DGIndex v1.5.8
    core.ffms2.Source is used on everything else
    if DGDecNV ever gets linux support I will test it out and add support. Until then its dead to me.
    """

    def __init__(self, file_path):
        if not hasattr(core, "d2v"):
            raise RuntimeError(
                "pvsfunc.PSourcer: Required plugin d2vsource for namespace 'd2v' not found. "
                "https://github.com/dwbuiten/d2vsource"
            )
        if not hasattr(core, "ffms2"):
            raise RuntimeError(
                "pvsfunc.PSourcer: Required plugin ffms2 for namespace 'ffms2' not found. "
                "https://github.com/FFMS/ffms2"
            )
        self.clip = None
        self.file_path = anti_file_prefix(file_path)
        self.file_type = get_mime_type(self.file_path)
        self.video_codec = get_video_codec(self.file_path)
        if self.file_type.startswith("image/"):
            self.video_codec = "IMAGE"
        self.sourcer = self.get_sourcer(self.video_codec)
        # sourcer preparations
        if self.sourcer == "core.d2v.Source":
            # make sure a d2v file for this video exists
            self.file_path = get_d2v(self.file_path)
        # load video to clip using sourcer
        self.clip = eval(self.sourcer)(self.file_path, **SOURCER_ARGS_MAP[self.sourcer])
        # set various props that the user may find helpful
        for k, v in [
            ("FilePath", self.file_path),
            ("Codec", self.video_codec),
            ("Sourcer", str(self.sourcer))
        ]:
            # why the fuck is there no SetFrameProps, come on...
            # +1: https://github.com/vapoursynth/vapoursynth/issues/559
            self.clip = core.std.SetFrameProp(self.clip, prop=f"PVS{k}", data=v)

    @staticmethod
    def get_sourcer(video_codec):
        """Get clip sourcer function based on video codec"""
        if video_codec not in CODEC_SOURCER_MAP:
            # default to FFMPEG-based sourcer for wide compatibility
            return "core.ffms2.Source"
        return CODEC_SOURCER_MAP[video_codec]
